from pathlib import Path
from collections.abc import Callable
from contextlib import contextmanager
from base64 import b64encode
from math import ceil

from ..logging import log_func_call  # , log_info
from .callback import qt_callback
from .widgets import GuiWidget, GuiWidgetView
from .qt import (
    QWidget, QToolButton, QSize, QIcon, QAction, QSizePolicy, QPainter,
    QBuffer, QByteArray, QSlider, Qt, QPushButton, QAbstractButton,
    QFontMetrics, QStyleOptionViewItem, QFont, QRect, QStyledItemDelegate,
    QTableWidget, QTableWidgetItem, QStyle, QTextLayout, QFIXED_MAX,
    QTextOption, QPointF,
)
# from .gui_app import get_gui_app

GuiWidgetParent = QWidget | GuiWidget | GuiWidgetView
TEXTWORDWRAP = Qt.TextWordWrap
ALIGNLEFT = Qt.AlignLeft


@log_func_call
def get_widget_parent_qtobj(parent: GuiWidgetParent) -> QWidget:
    """Get the Qt object of the parent widget."""
    if isinstance(parent, QWidget):
        return parent
    elif isinstance(parent, GuiWidget):
        return parent.gui_parent.qtobj
    elif isinstance(parent, GuiWidgetView):
        return parent.qtobj
    else:
        raise TypeError(f"Unsupported parent type: {type(parent)}")


@log_func_call
def load_icon(icon_path: str | Path) -> QIcon:
    return QIcon(Path(icon_path).as_posix())


# @log_func_call
# def create_toolbtn(parent: GuiWidgetParent,
#                    callback: Callable = None,
#                    sustain: bool = False, sus_repeat_interval_ms: int = 33,
#                    sus_delay_ms: int = 0, toggleable: bool = False,
#                    toggle_depressed: bool = False, enabled: bool = True):
#     button = QToolButton(get_widget_parent_qtobj(parent))
#     button.setEnabled(enabled)
#     if sustain:
#         button.setAutoRepeat(True)
#         button.setAutoRepeatInterval(sus_repeat_interval_ms)
#         button.setAutoRepeatDelay(sus_delay_ms)

#     if toggleable:
#         button.setCheckable(True)
#         button.setChecked(toggle_depressed)

#     if callback:
#         signal = button.toggled if toggleable else button.clicked
#         signal.connect(callback)

#     return button


# @log_func_call
# def create_icon_toolbtn(parent: GuiWidgetParent, size: QSize,
#                         icon: QIcon | str | Path,
#                         callback: Callable = None,
#                         sustain: bool = False,
#                         sus_repeat_interval_ms: int = 33,
#                         sus_delay_ms: int = 0, toggleable: bool = False,
#                         toggle_depressed: bool = False,
#                         enabled: bool = True):
#     if not isinstance(icon, QIcon):
#         icon = load_icon(icon)

#     button = create_toolbtn(parent, callback, sustain,
#                             sus_repeat_interval_ms,
#                             sus_delay_ms, toggleable, toggle_depressed,
#                             enabled)
#     button.setIcon(icon)
#     button.setIconSize(size)
#     button.setMinimumHeight(calculate_min_height(button, size))
#     _debug_dump_icon(icon, size, name=f"toolbtn_{icon}")
#     return button


# @log_func_call
# def create_text_toolbtn(parent: GuiWidgetParent, text: str,
#                         callback: Callable = None,
#                         sustain: bool = False,
#                         sus_repeat_interval_ms: int = 33,
#                         sus_delay_ms: int = 0, toggleable: bool = False,
#                         toggle_depressed: bool = False,
#                         enabled: bool = True):
#     button = create_toolbtn(parent, callback, sustain,
#                             sus_repeat_interval_ms,
#                             sus_delay_ms, toggleable, toggle_depressed,
#                             enabled)
#     button.setText(text)
#     return button


@log_func_call
def create_slider(parent: GuiWidgetParent, min_value: int,
                  max_value: int, value: int, callback: Callable = None):
    slide = QSlider(get_widget_parent_qtobj(parent))
    slide.setMinimum(min_value)
    slide.setMaximum(max_value)
    slide.setValue(value)
    # slide.setOrientation(orientation)
    # slide.setEnabled(enabled)

    if callback:
        slide.valueChanged.connect(qt_callback(callback))

    return slide


@log_func_call
def show_toolbtn_icon_and_text(btn: QToolButton):
    btn.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)


@log_func_call
def create_action(parent: GuiWidgetParent, text: str = "",
                  icon: QIcon | str | Path = None,
                  callback: Callable = None, enabled: bool = True,
                  tooltip: str = None, checkable: bool = False):
    action = QAction(get_widget_parent_qtobj(parent))
    if text:
        action.setIconText(text)

    if tooltip:
        action.setToolTip(tooltip)

    if icon:
        if not isinstance(icon, QIcon):
            icon = load_icon(icon)

        action.setIcon(icon)

    if callback:
        action.triggered.connect(qt_callback(callback))

    if checkable:
        action.setCheckable(True)

    action.setEnabled(enabled)
    return action


@log_func_call
def set_widget_sizepolicy_h_expanding(w: QWidget):
    w.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)


@log_func_call
def set_widget_sizepolicy_v_expanding(w: QWidget):
    w.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)


@log_func_call
def set_widget_sizepolicy_hv_expanding(w: QWidget):
    w.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)


@log_func_call
def create_toolbar_expanding_spacer():
    spacer = QWidget()
    set_widget_sizepolicy_hv_expanding(spacer)
    return spacer


@contextmanager
def painter_context(painter: QPainter):

    painter.save()
    try:
        yield painter
    finally:
        painter.restore()


@log_func_call
def qicon_to_data_uri(icon: QIcon, size: QSize) -> str:
    """Convert QIcon to data URI"""
    # Get pixmap from icon
    pixmap = icon.pixmap(size)

    # Convert to PNG bytes
    byte_array = QByteArray()
    buffer = QBuffer(byte_array)
    buffer.open(QBuffer.WriteOnly)
    pixmap.save(buffer, "PNG")

    # Encode as base64 data URI
    png_data = b64encode(byte_array.data()).decode('ascii')
    return f"data:image/png;base64,{png_data}"


@log_func_call
def create_button(parent: GuiWidgetParent,
                  callback: Callable = None,
                  sustain: bool = False, sus_repeat_interval_ms: int = 33,
                  sus_delay_ms: int = 0, toggleable: bool = False,
                  toggle_depressed: bool = False, enabled: bool = True):
    button = QPushButton(get_widget_parent_qtobj(parent))
    button.setEnabled(enabled)
    if sustain:
        button.setAutoRepeat(True)
        button.setAutoRepeatInterval(sus_repeat_interval_ms)
        button.setAutoRepeatDelay(sus_delay_ms)

    if toggleable:
        button.setCheckable(True)
        button.setChecked(toggle_depressed)

    if callback:
        signal = button.toggled if toggleable else button.clicked
        signal.connect(qt_callback(callback))

    return button


@log_func_call
def create_icon_button(parent: GuiWidgetParent, size: QSize,
                       icon: QIcon | str | Path,
                       callback: Callable = None,
                       caption: str = None,
                       sustain: bool = False,
                       sus_repeat_interval_ms: int = 33,
                       sus_delay_ms: int = 0, toggleable: bool = False,
                       toggle_depressed: bool = False, enabled: bool = True):
    if not isinstance(icon, QIcon):
        icon = load_icon(icon)

    button = create_button(parent, callback, sustain, sus_repeat_interval_ms,
                           sus_delay_ms, toggleable, toggle_depressed,
                           enabled)
    button.setIcon(icon)
    button.setIconSize(size)
    button.setMinimumHeight(calculate_min_height(button, size))
    # optional debug dump
    _debug_dump_icon(icon, size, name=caption)
    if caption:
        button.setText(caption)
    return button


@log_func_call
def calculate_min_height(button: QAbstractButton, icon_size: QSize):
    """Calculate the minimum height for a button to accommodate its icon."""
    # Ensure the button is tall enough to vertically center the icon.
    # If the icon size is set explicitly in code (e.g. 32x32), the
    # QPushButton can be shorter than the icon which makes the icon
    # appear visually off-center. Set a sensible minimum height that
    # respects the icon height plus a small vertical padding.
    return max(button.sizeHint().height(), icon_size.height() + 6)


@log_func_call
def _debug_dump_icon(icon: QIcon, size: QSize, name: str = None) -> None:
    from ..app import PyRandyOSApp
    if PyRandyOSApp.get('debug_dump_icons', False):
        tmpdir = PyRandyOSApp.mkdir_temp('icons')
        fname = f"icon_dump_{name or 'icon'}.png"
        pix = icon.pixmap(size)
        pix.save(str(tmpdir/fname), "PNG")


@log_func_call
def wrap_text_to_width(text: str, bbox_callback: Callable[[str], QRect],
                       width: int):
    "Wrap text to fit within a specified width using the given font metrics."
    # gui = get_gui_app()
    # log_info(f"DPI: {gui.get_dpi()}")
    # log_info(f"DPI Scale: {gui.get_dpi_scale()}")
    # log_info(f"Window DPI: {gui.windows[0].gui_view.get_dpi()}")

    # compute column width
    if not text or width <= 0:
        return text

    @log_func_call
    def calc_width(s: str) -> int:
        return bbox_callback(s).width()

    # as-is original width
    test_width = calc_width(text)
    if test_width <= width:
        return text

    # Split text by existing newlines first, then wrap each line
    original_lines = text.splitlines()
    wrapped = []
    append_wrapped = wrapped.append
    for origline in original_lines:
        if not origline.strip():
            # Empty line - preserve it
            append_wrapped("")
            continue

        # Check if this line fits without wrapping
        test_width = calc_width(origline)
        if test_width <= width:
            append_wrapped(origline)
            continue

        # Apply word wrapping to this line
        curline = ""
        words = origline.split()

        @log_func_call
        def popword():
            return words.pop(0) if words else None

        w = popword()  # we already handled blank line, so should work
        while w:
            # Calculate width of current line + new word
            test_line = curline + (" " if curline else "") + w
            test_width = calc_width(test_line)
            # if it fits, i sits...try adding another word
            if test_width <= width:
                curline = test_line
                w = popword()
                continue
            # w pushed us over the limit, so break the line and start over
            if curline:
                append_wrapped(curline)
                curline = ""
                continue
            # we got a long single word, break by character
            w2 = ""
            for c in w:
                test_line = w2 + c
                test_width = calc_width(test_line)
                if test_width > width:
                    # c put us over the limit, flush w2
                    if w2:
                        append_wrapped(w2)
                        w2 = c
                    else:
                        # single char is too much, but we need at least 1
                        append_wrapped(c)
                        w2 = ""
                else:
                    w2 = test_line

            # curline is still blank if we got here because we are in
            # single word mode
            w = w2 if w2 else popword()

        if curline:
            append_wrapped(curline)

    return "\n".join(wrapped)


@log_func_call
def calc_fontmetrics_bbox(font_metrics: QFontMetrics, s: str) -> QRect:
    return font_metrics.boundingRect(s)


@log_func_call
def get_styled_text_bbox(widget: QWidget, font: QFont, text: str):
    style_option = QStyleOptionViewItem()
    style_option.font = font
    style_option.fontMetrics = QFontMetrics(font)
    # Get the style from your table widget
    style = widget.style()
    rect = style.itemTextRect(style_option.fontMetrics, QRect(), ALIGNLEFT,
                              True, text)
    return rect


@log_func_call
def get_table_item_text_bbox(table: QTableWidget, item: QTableWidgetItem,
                             text: str) -> QRect:
    # adapted from internal Qt C++ source code for resizeColumnToContents()
    # and all the many other places it calls, consolidated here.
    table.ensurePolished()

    # collect necessary info from table
    style = table.style()
    tblopt = table.viewOptions()

    # get the table index for the item
    idx = table.indexFromItem(item)

    # get the item delegate from the table, which we need to create a
    # QStyleOptionViewItem for the item.  We start by copying the table default
    # styles, and then use initStyleOption() to apply the item-specific
    # styles to the option.
    delegate: QStyledItemDelegate = table.itemDelegate(idx)
    itemopt = QStyleOptionViewItem(tblopt)
    delegate.initStyleOption(itemopt, idx)

    # get the text width margin for the item
    text_margin = style.pixelMetric(QStyle.PM_FocusFrameHMargin, itemopt,
                                    table) + 1

    # create a text layout to measure the text
    text_layout = QTextLayout(delegate.displayText(text, itemopt.locale),
                              itemopt.font)

    # set text options
    text_option = QTextOption()
    wrap_text = int(itemopt.features & QStyleOptionViewItem.WrapText)
    text_option.setWrapMode(QTextOption.WordWrap if wrap_text
                            else QTextOption.ManualWrap)
    text_option.setTextDirection(itemopt.direction)
    text_option.setAlignment(QStyle.visualAlignment(itemopt.direction,
                                                    itemopt.displayAlignment))
    text_layout.setTextOption(text_option)

    # measure the text
    height = 0
    width = 0
    text_layout.beginLayout()
    line = text_layout.createLine()
    while line.isValid():
        line.setLineWidth(QFIXED_MAX)
        line.setPosition(QPointF(0, height))
        height += line.height()
        width = max(width, line.naturalTextWidth())
        line = text_layout.createLine()

    text_layout.endLayout()
    return QRect(0, 0, ceil(width) + 2 * text_margin,
                 ceil(height) + 2 * text_margin)

    # style.sizeFromContents(QStyle.CT_ItemViewItem, itemopt, QSize(), table)
    # rect = style.itemTextRect(itemopt.fontMetrics, QRect(), ALIGNLEFT,
    #                           True, text)
    # return rect


@log_func_call
def get_table_item_usable_rect(table: QTableWidget, item: QTableWidgetItem):
    table.ensurePolished()

    # collect necessary info from table
    style = table.style()
    tblopt = table.viewOptions()

    # get the table index for the item
    idx = table.indexFromItem(item)

    # get the item delegate from the table, which we need to create a
    # QStyleOptionViewItem for the item.  We start by copying the table default
    # styles, and then use initStyleOption() to apply the item-specific
    # styles to the option.
    itemopt = QStyleOptionViewItem(tblopt)
    delegate: QStyledItemDelegate = table.itemDelegate(idx)
    delegate.initStyleOption(itemopt, idx)

    # initialize the rect in the itemopt with the full visual rect of the item
    visualrect = table.visualItemRect(item)
    itemopt.rect = visualrect

    # compute the content rect from the style
    contentrect = style.subElementRect(QStyle.SE_ItemViewItemText, itemopt,
                                       table)
    return contentrect


# def test_width_calc(table: QTableWidget, item: QTableWidgetItem,
#                     width: int, text: str):
#     # based on QItemDelegate::drawDisplay()
#     rect = QRect(0, 0, width, 200)  # arbitrary large height
#     tblopt = table.viewOptions()
#     idx = table.indexFromItem(item)
#     delegate: QStyledItemDelegate = table.itemDelegate(idx)
#     option = QStyleOptionViewItem(tblopt)
#     delegate.initStyleOption(option, idx)
#     textOption = QTextOption()
#     textLayout = QTextLayout()

#     opt = QStyleOptionViewItem(option)
#     widget = table
#     style = widget.style()
#     textMargin = style.pixelMetric(QStyle.PM_FocusFrameHMargin, None,
#                                    widget) + 1
#     # remove width padding
#     textRect = rect.adjusted(textMargin, 0, -textMargin, 0)
#     wrapText = int(opt.features & QStyleOptionViewItem.WrapText)
#     textOption.setWrapMode(QTextOption.WordWrap if wrapText
#                            else QTextOption.ManualWrap)
#     textOption.setTextDirection(option.direction)
#     textOption.setAlignment(QStyle.visualAlignment(option.direction,
#                                                    option.displayAlignment))
#     textLayout.setTextOption(textOption)
#     textLayout.setFont(option.font)
#     textLayout.setText(text)

#     lineWidth = textRect.width()

#     height = 0
#     widthUsed = 0
#     textLayout.beginLayout()
#     line = textLayout.createLine()
#     while line.isValid():
#         line.setLineWidth(lineWidth)
#         line.setPosition(QPointF(0, height))
#         height += line.height()
#         widthUsed = max(widthUsed, line.naturalTextWidth())
#         line = textLayout.createLine()

#     textLayout.endLayout()
#     return lineWidth, widthUsed



# The MIT License
#
# Copyright (c) 2015 The Spyder development team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import TYPE_CHECKING

from ....logging import log_func_call, DEBUGLOW2
from ...callback import qt_callback
from ....utils.hash import TupleHashMixin
from ...qt import QTimer, QRect, QRectF, QWidget, QPainter
if TYPE_CHECKING:
    from .icon import IconLayer


class IconAnimation(TupleHashMixin):
    """
    Base class for icon animations.
    This class provides a common interface for animations that can be applied
    to icons.
    """
    def __init__(self, parent_widget: QWidget):
        self.parent_widget = parent_widget

    def setup(self, icon_painter: 'IconLayer', painter: QPainter,
              rect: QRect | QRectF):
        raise NotImplementedError("Subclasses should implement this method")

    def start(self):
        raise NotImplementedError("Subclasses should implement this method")

    def stop(self):
        raise NotImplementedError("Subclasses should implement this method")

    def as_tuple(self):
        """
        Returns a tuple representation of the animation instance.
        This is used for hashing and equality checks.
        """
        return (self.__class__.__qualname__, self.parent_widget)


class IconSpin(IconAnimation):
    @log_func_call
    def __init__(self, parent_widget: QWidget, interval: int = 10,
                 step: int = 1, autostart: bool = True):
        super().__init__(parent_widget)
        self.interval = interval
        self.step = step
        self.autostart = autostart

        self.info = {}

    def as_tuple(self):
        return super().as_tuple() + (
            self.interval,
            self.step,
            self.autostart,
        )

    @log_func_call(DEBUGLOW2, trace_only=True)
    def _update(self):
        if self.parent_widget in self.info:
            timer, angle, step = self.info[self.parent_widget]

            if angle >= 360:
                angle = 0

            angle += step
            self.info[self.parent_widget] = timer, angle, step
            self.parent_widget.update()

    @log_func_call
    def setup(self, icon_painter: 'IconLayer', painter: QPainter,
              rect: QRect | QRectF):
        if self.parent_widget not in self.info:
            timer = QTimer(self.parent_widget)
            timer.timeout.connect(qt_callback(self._update))
            self.info[self.parent_widget] = [timer, 0, self.step]
            if self.autostart:
                timer.start(self.interval)
        else:
            timer, angle, self.step = self.info[self.parent_widget]
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            painter.rotate(angle)
            painter.translate(-x_center, -y_center)

    @log_func_call
    def start(self):
        if self.parent_widget in self.info:
            timer: QTimer = self.info[self.parent_widget][0]
            timer.start(self.interval)

    @log_func_call
    def stop(self):
        if self.parent_widget in self.info:
            timer: QTimer = self.info[self.parent_widget][0]
            timer.stop()


class IconPulse(IconSpin):
    @log_func_call
    def __init__(self, parent_widget: QWidget, autostart: bool = True):
        super().__init__(parent_widget, interval=300, step=45,
                         autostart=autostart)

    def as_tuple(self):
        return super().as_tuple() + (
            self.autostart,
        )



# The MIT License
#
# Copyright (c) 2015 The Spyder development team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import TYPE_CHECKING
from importlib import import_module
from weakref import WeakValueDictionary

from ....logging import log_func_call, DEBUGLOW2  # noqa: E402
from ....utils.hash import TupleHashMixin  # noqa: E402
from ...gui_app import get_gui_app
from ...qt import (
    QRect, Qt, QSizeF, QRectF, QPointF, QSize, QPoint, QPainter, QColor,
    QTransform, QFont, QRawFont, QImage, QIcon, QIconEngine, QPixmap, QPalette,
    QGlyphRun,
)
from ...utils import painter_context  # noqa: E402
from .animation import IconAnimation  # noqa: E402
from .sources import THIRDPARTY_FONTSPEC  # noqa: E402
if TYPE_CHECKING:
    from .font import IconFont

_NOHINT = QFont.PreferNoHinting
_DEFHINT = QFont.PreferDefaultHinting

IconCache = WeakValueDictionary[str, QIcon]
PaletteType = QPalette.ColorRole | tuple[QPalette.ColorGroup,
                                         QPalette.ColorRole]
RgbTriple = tuple[int, int, int]
RgbaQuad = tuple[int, int, int, int]
ColorType = str | QColor | PaletteType | RgbTriple | RgbaQuad


class IconLayer(TupleHashMixin):
    def as_tuple(self):
        anim = self.animation
        return (
            self.font._SPECNAME,
            self.glyph,
            self._orig_color,
            self.x,
            self.y,
            anim.as_tuple() if anim else None,
            self.scale,
            self.opacity,
            self.rotation,
            self.hflip,
            self.vflip,
            self.draw,
            self._alpha,
            self._is_disabled_icon,
        )

    @log_func_call(DEBUGLOW2, trace_only=True)
    def __init__(
        self,
        font: 'IconFont | type[IconFont] | str',
        glyph: str | int = None,
        color: ColorType = None,
        x: int = 0,
        y: int = 0,
        animation: IconAnimation | tuple = None,
        scale: float = 1.0,
        opacity: float = 1.0,
        rotation: float = 0.0,
        hflip: bool = False,
        vflip: bool = False,
        draw: str = None,

        # these are only used to precompute color:
        alpha: int = None,
        is_disabled_icon: bool = False,

        *,
        glyph_name: str = None,  # not saved, just used to set glyph
    ):
        if isinstance(font, str):
            font = THIRDPARTY_FONTSPEC[font].get_font_class()

        self.font = font
        if glyph and glyph_name:
            raise ValueError(
                "Cannot specify both 'glyph' and 'glyph_name'. "
                "Use 'glyph_name' to set the glyph from the font's charmap."
            )
        if not glyph:
            if glyph_name:
                glyph = font.get_codepoint_by_name(glyph_name)
                if not glyph:
                    raise ValueError(
                        f"Glyph name '{glyph_name}' not found in "
                        f"font {font._SPECNAME}."
                    )
            else:
                raise ValueError(
                    "Either 'glyph' or 'glyph_name' must be specified."
                )
        self.glyph = glyph if isinstance(glyph, str) else chr(glyph)

        self._orig_color = color
        if not color:
            color = ((QPalette.Disabled, QPalette.WindowText)
                     if is_disabled_icon else QPalette.WindowText)

        if not self.is_palette_color(color) and not isinstance(color, QColor):
            color = QColor(color) if isinstance(color, str) else QColor(*color)

        self.color = color
        self.x = x
        self.y = y

        if isinstance(animation, tuple):
            qualname: str = animation[0]
            qualnameparts = qualname.split('.')
            mod = import_module(qualnameparts[:-1])
            animation = getattr(mod, qualnameparts[-1])(*animation[1:])

        self.animation = animation
        self.scale = scale
        self.opacity = opacity
        self.rotation = rotation
        self.hflip = hflip
        self.vflip = vflip
        self.draw = draw or ('path' if animation else 'text')
        self._alpha = alpha
        self._is_disabled_icon = is_disabled_icon

    @staticmethod
    def is_palette_color(color: ColorType):
        return (isinstance(color, QPalette.ColorRole) or
                (isinstance(color, tuple)
                 and isinstance(color[0], QPalette.ColorGroup)
                 and isinstance(color[1], QPalette.ColorRole)))

    def get_color(self):
        color = self.color
        if self.is_palette_color(color):
            app = get_gui_app()
            if app:
                palette = app.qtobj.palette()
                color = palette.color(*(color if isinstance(color, tuple)
                                        else (color,)))
            else:
                color = (QColor(150, 150, 150) if self._is_disabled_icon
                         else QColor(50, 50, 50))

        alpha = self._alpha
        if alpha:
            color.setAlpha(alpha)

        return color

    def paint(self, painter: QPainter, rect: QRect):
        font = self.font
        with painter_context(painter):
            painter.setPen(self.get_color())

            anim = self.animation
            if anim:
                anim.setup(font, painter, rect)

            rect = QRect(rect)
            rect.translate(self.x*rect.width(), self.y*rect.height())
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            transform = QTransform()
            transform.translate(x_center, y_center)
            if self.vflip:
                transform.scale(1, -1)

            if self.hflip:
                transform.scale(-1, 1)

            rotate = self.rotation
            if rotate:
                transform.rotate(rotate)

            transform.translate(-x_center, -y_center)
            painter.setTransform(transform)

            painter.setOpacity(self.opacity)
            draw = self.draw
            rawfont = self.get_rawfont(rect) if draw != 'text' else None
            done = False
            if rawfont:
                if draw == 'path':
                    done = self.draw_path(painter, rect, rawfont)

                elif draw == 'glyphrun':
                    done = self.draw_glyphrun(painter, rect, rawfont)

                elif draw == 'image':
                    done = self.draw_image(painter, rect, rawfont)

            if draw == 'text' or not done:
                done = self.draw_text(painter, rect)

    def get_draw_size(self, rect: QRect):
        # A 16 pixel-high icon yields a font size of 14, which is pixel
        # perfect for font-awesome. 16 * 0.875 = 14
        # The reason why the glyph size is smaller than the icon size is to
        # account for font bearing.
        return round(0.875*rect.height()*self.scale)

    def draw_text(self, painter: QPainter, rect: QRect):
        font = self.font.get_font(self.get_draw_size(rect))
        if self.animation:
            font.setHintingPreference(_NOHINT)
        painter.setFont(font)
        painter.drawText(rect, Qt.AlignCenter | Qt.AlignVCenter, self.glyph)
        return True

    def get_rawfont(self, rect: QRect):
        hint = (_NOHINT if (self.animation and self.draw == "glyphrun")
                else _DEFHINT)
        rawfont = self.font.get_rawfont(self.get_draw_size(rect), hint)
        if rawfont.fontTable("glyf"):
            return rawfont

    def prep_rawfont_paint(self, painter: QPainter, rect: QRect,
                           rawfont: QRawFont):
        glyph = rawfont.glyphIndexesForString(self.glyph)[0]
        advance = rawfont.advancesForGlyphIndexes((glyph,))[0]
        ascent = rawfont.ascent()
        size = QSizeF(abs(advance.x()), ascent + rawfont.descent())
        painter.translate(QRectF(rect).center())
        painter.translate(-size.width() / 2, -size.height() / 2)
        return glyph, ascent

    def draw_path(self, painter: QPainter, rect: QRect, rawfont: QRawFont):
        glyph, ascent = self.prep_rawfont_paint(painter, rect, rawfont)
        path = rawfont.pathForGlyph(glyph)
        path.translate(0, ascent)
        path.setFillRule(Qt.WindingFill)
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.fillPath(path, painter.pen().color())
        return True

    def draw_glyphrun(self, painter: QPainter, rect: QRect, rawfont: QRawFont):
        if not QGlyphRun:
            return False

        glyph, ascent = self.prep_rawfont_paint(painter, rect, rawfont)
        glyphrun = QGlyphRun()
        glyphrun.setRawFont(rawfont)
        glyphrun.setGlyphIndexes((glyph,))
        glyphrun.setPositions((QPointF(0, ascent),))
        painter.drawGlyphRun(QPointF(0, 0), glyphrun)
        return True

    def draw_image(self, painter: QPainter, rect: QRect, rawfont: QRawFont):
        glyph, ascent = self.prep_rawfont_paint(painter, rect, rawfont)
        image = rawfont.alphaMapForGlyph(
            glyph, QRawFont.PixelAntialiasing
        ).convertToFormat(QImage.Format_ARGB32_Premultiplied)
        painter2 = QPainter(image)
        painter2.setCompositionMode(QPainter.CompositionMode_SourceIn)
        painter2.fillRect(image.rect(), painter.pen().color())
        painter2.end()
        brect = rawfont.boundingRect(glyph)
        brect.translate(0, ascent)
        painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
        painter.drawImage(brect.topLeft(), image)
        return True


IconLayerStack = list[IconLayer]  # A stack of IconLayer objects
IconLayerOrStack = IconLayer | IconLayerStack


def ensure_icon_layer_stack(layers: IconLayerOrStack) -> IconLayerStack:
    return [layers] if isinstance(layers, IconLayer) else layers


class IconStateSpec(TupleHashMixin):
    def __init__(self, normal: IconLayerOrStack,
                 active: IconLayerOrStack = None,
                 selected: IconLayerOrStack = None,
                 disabled: IconLayerOrStack = None,
                 ):
        normal = ensure_icon_layer_stack(normal)
        active = ensure_icon_layer_stack(active or normal)
        selected = ensure_icon_layer_stack(selected or active)

        if not disabled:
            disabled = [IconLayer(x.font,
                                  *x.as_tuple()[1:-1],
                                  is_disabled_icon=True)
                        for x in normal]

        self.normal = normal
        self.active = active
        self.selected = selected
        self.disabled = ensure_icon_layer_stack(disabled)

    def get_layers_for_mode(self, mode: QIcon.Mode):
        if mode == QIcon.Mode.Active:
            return self.active
        elif mode == QIcon.Mode.Selected:
            return self.selected
        elif mode == QIcon.Mode.Disabled:
            return self.disabled
        else:
            return self.normal

    def as_tuple(self):
        return (
            tuple(layer.as_tuple() for layer in self.normal),
            tuple(layer.as_tuple() for layer in self.active),
            tuple(layer.as_tuple() for layer in self.selected),
            tuple(layer.as_tuple() for layer in self.disabled),
        )


class IconSpec(TupleHashMixin):
    _cache: IconCache = WeakValueDictionary()

    def __init__(self, on: IconStateSpec, off: IconStateSpec = None):
        self.on = on
        self.off = off or on

    def get_layers_for_state_mode(self, state: QIcon.State, mode: QIcon.Mode):
        if state == QIcon.State.Off:
            return self.off.get_layers_for_mode(mode)
        else:
            return self.on.get_layers_for_mode(mode)

    def as_tuple(self):
        return (
            self.on.as_tuple(),
            self.off.as_tuple(),
        )

    @log_func_call(DEBUGLOW2, trace_only=True)
    def icon(self):
        cache = self._cache
        cache_key = self.as_tuple()
        icon = cache.get(cache_key)
        if not icon:
            icon = QIcon(FontIconEngine(self))
            cache[cache_key] = icon

        return icon

    @classmethod
    @log_func_call(DEBUGLOW2, trace_only=True)
    def generate_iconspec(cls, font: 'IconFont | type[IconFont] | str',
                          glyph_name: str = None, glyph: str | int = None):
        """
        Generate a simple IconSpec object with default settings for the given
        font and glyph.  Note that one of either `glyph_name` or `glyph` must
        be provided or a ValueError will be raised.

        Note that if `pyrandyos.gui.icons.iconfont.init_iconfonts()` has not
        been called prior to this call, "name" string values cannot be
        resolved.  Only the IconFont subclass for the font or an instance
        thereof, and the explicit "char" string or integer codepoint, may be
        passed to this method on early imports.
        Otherwise, a IconFontNotInitializedError will be raised.

        Args:
            font (IconFont | type[IconFont] | str):
                The font from which the glyph will be retrieved.  This may be
                the fontspec label as a string, an instance of its `IconFont`
                subclass, or even just the font's `IconFont` subclass itself.
            glyph_name (str, optional): The detailed name string of the
                character as listed in the character map for the font.
                Defaults to None.
            glyph (str | int, optional): Either the numeric codepoint or the
                actual Unicode character to use from the font.
                Defaults to None.

        Returns:
            IconSpec: icon spec for the given

        Raises:
            IconFontNotInitializedError: the given font has not yet been loaded
            ValueError: incorrect glyph arguments provided
        """
        return cls(IconStateSpec(IconLayer(font, glyph,
                                           glyph_name=glyph_name)))


class FontIconEngine(QIconEngine):
    def __init__(self, iconspec: IconSpec):
        super().__init__()
        self.iconspec = iconspec

    def paint(self, painter: QPainter, rect: QRect, mode: QIcon.Mode,
              state: QIcon.State):
        iconspec = self.iconspec
        for layer in iconspec.get_layers_for_state_mode(state, mode):
            layer.paint(painter, rect)

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State):
        pm = QPixmap(size)
        pm.fill(Qt.transparent)
        self.paint(QPainter(pm), QRect(QPoint(0, 0), size), mode, state)
        return pm

    def iconName(self):
        return str(self.iconspec.as_tuple())



from ..logging import log_func_call, DEBUGLOW2, DEBUG

from . import GuiPresenter
from .widgets import GuiWindowLikeView
from .qt import (
    QSplashScreen, QPixmap, QProgressBar, QLabel, Qt, QMouseEvent, QRect,
    QPalette,
)
from .loadstatus import LOAD_STEP_REGISTRY


class GuiSplashScreen(GuiPresenter['GuiSplashScreenView']):
    def __init__(self, pixmap: QPixmap = None, *view_args,
                 text: str = "Loading...", **view_kwargs):
        super().__init__(None, pixmap, *view_args, text=text, **view_kwargs)

    def create_gui_view(self, pixmap: QPixmap = None, *view_args,
                        text: str = "Loading...", **view_kwargs):
        return GuiSplashScreenView(self, pixmap, *view_args, text=text,
                                   **view_kwargs)

    def mousePressEvent(self, event: QMouseEvent):
        # Prevent the splash from being closed by mouse click
        pass

    def post_show_init(self):
        "Any initialization that should be executed after the window is shown"
        pass

    def show(self):
        self.gui_view.show()

    @log_func_call(DEBUGLOW2, trace_only=True)
    def set_progress(self, *, value: int = None, message: str = None,
                     process_events: bool = True):
        view = self.gui_view
        progress = view.progress
        mymax = progress.maximum()
        refresh = mymax // 200 or 1
        process_events = process_events and ((message and value is None)
                                             or (value
                                                 and value % refresh == 0)
                                             or value == mymax
                                             or value == 0
                                             )

        if value is not None:
            progress.setValue(value)

        if message:
            view.label.setText(message)
            # self.qtobj.showMessage(message,
            #                         Qt.AlignBottom | Qt.AlignHCenter,
            #                         Qt.white)

        if process_events:
            self.qt_app.processEvents()


class GuiSplashScreenView(GuiWindowLikeView[GuiSplashScreen, QSplashScreen]):
    @log_func_call(DEBUG)
    def __init__(self, presenter: GuiSplashScreen, pixmap: QPixmap = None,
                 *qtobj_args, text: str = "Loading...", **qtobj_kwargs):
        GuiWindowLikeView.__init__(self, text, presenter, pixmap, *qtobj_args,
                                   **qtobj_kwargs)
        qtobj = self.qtobj
        qtobj.mousePressEvent = presenter.mousePressEvent
        self.center_window_in_current_screen()

        self.rect = pixmap.rect() if pixmap else QRect(0, 0, 400, 400)
        self.gridsize = 20
        self.default_text = text
        self.create_pb()
        self.create_label()

    @log_func_call(DEBUGLOW2, trace_only=True)
    def create_qtobj(self, pixmap: QPixmap = None,
                     flags: Qt.WindowFlags = Qt.WindowStaysOnTopHint,
                     *args, **kwargs):
        return QSplashScreen(pixmap or QPixmap(), flags, *args, **kwargs)

    def create_pb(self):
        qtobj = self.qtobj
        gridsize = self.gridsize
        rect = self.rect

        pb = QProgressBar(qtobj)
        pb.setGeometry(gridsize, round((rect.height() + gridsize)*0.75),
                       rect.width() - 2*gridsize, gridsize)
        pb.setRange(0, len(LOAD_STEP_REGISTRY) or 1)
        pb.setValue(0)
        pb.setStyleSheet('''
            QProgressBar {
                background-color: #19232D;
                border: 1px solid #455364;
                color: #DFE1E2;
                border-radius: 4px;
                text-align: center;
            }

            QProgressBar:disabled {
                background-color: #19232D;
                border: 1px solid #455364;
                color: #788D9C;
                border-radius: 4px;
                text-align: center;
            }

            QProgressBar::chunk {
                background-color: #346792;
                color: #19232D;
                border-radius: 4px;
            }

            QProgressBar::chunk:disabled {
                background-color: #26486B;
                color: #788D9C;
                border-radius: 4px;
            }
        ''')
        self.progress = pb

    def create_label(self):
        qtobj = self.qtobj
        gridsize = self.gridsize
        rect = self.rect
        text = self.default_text

        lbl = QLabel(qtobj)
        lbl.setGeometry(gridsize, round((rect.height() - gridsize)*0.75),
                        rect.width() - 2*gridsize, gridsize)
        lbl.setAlignment(Qt.AlignCenter)
        lbl.setText(text)
        palette = lbl.palette()
        palette.setColor(QPalette.Window, lbl.palette().color(QPalette.Window))
        lbl.setAutoFillBackground(True)
        lbl.setPalette(palette)
        lbl.setStyleSheet('''
            QLabel {
                padding: 2px 8px;
                background: #444a5a;
                color: #fff;
                border-radius: 4px;
            }
        ''')
        self.label = lbl



from ....logging import log_func_call
from ....app import PyRandyOSApp
from ...loadstatus import load_status_step

from .font import IconFont  # noqa: F401
from .sources import THIRDPARTY_FONTSPEC  # noqa: F401
from .fontspec import ICON_ASSETS_DIR
from .icon import IconSpec, IconStateSpec, IconLayer  # noqa: F401


@load_status_step("Loading icon fonts")
@log_func_call
def init_iconfonts(use_tmpdir: bool = True, do_import: bool = True):
    tmpdir = PyRandyOSApp.mkdir_temp() if use_tmpdir else ICON_ASSETS_DIR
    for fontmod, fontspec in THIRDPARTY_FONTSPEC.items():
        fontspec.initialize(fontmod, tmpdir)
        if do_import:
            fontspec.import_font()



from datetime import datetime
from sqlite3 import Connection
from contextlib import contextmanager


from ..logging import log_func_call

GET_TABLES_QUERY = "select name from sqlite_master where type='table'"


@log_func_call
def get_tables(db: Connection):
    """
    Get all table names in the database.
    """
    return tuple(row[0] for row in db.execute(GET_TABLES_QUERY).fetchall())


@log_func_call
def check_table_exists(db: Connection, table_name: str):
    """
    Check if a table exists in the database.
    """
    return table_name in get_tables(db)


@log_func_call
def get_table_fields(db: Connection, table_name: str):
    """
    Get all fields from the specified table of the database.
    """
    if not check_table_exists(db, table_name):
        return False

    query = f"pragma table_info({table_name!r})"
    return tuple(row[1] for row in db.execute(query).fetchall())


@log_func_call
def execute_select(db: Connection, table_name: str,
                   fields: str | list[str] = None, conditions: str = None,
                   expansions: tuple[str] = ()):
    """
    Execute select for specified fields from a table in the database.
    """
    if not check_table_exists(db, table_name):
        raise ValueError(f"Table '{table_name}' does not exist in "
                         "the database.")

    if fields:
        if isinstance(fields, str):
            fields = (fields,)

        valid_fields = get_table_fields(db, table_name)
        if not all(f in valid_fields for f in fields):
            raise ValueError("One or more fields do not exist in "
                             f"table '{table_name}'.")
        field_list = ', '.join(fields)
    else:
        field_list = '*'

    query = (f"select {field_list} from {table_name!r} "
             f"{' ' + conditions if conditions else ''}")
    # log_debug(f"Executing query: {query} with expansions: {expansions}")

    return db.execute(query, expansions)


@log_func_call
def sql_timestamp_to_datetime(ts: str):
    return datetime.fromisoformat(ts.replace('Z', '+00:00')) if ts else None


@contextmanager
@log_func_call
def sqlite_context(cxn: Connection):
    try:
        yield cxn
    except Exception:
        raise
    else:
        cxn.commit()
    finally:
        cxn.close()



"""
This module is intended for internal use only.  No user documentation is
provided at this time.  Use at your own discretion.
"""

from pathlib import Path
from ctypes import POINTER, c_ssize_t, c_ulonglong, c_ubyte

ErrCodeLookup = dict[int, str]
FuncErrMsgDict = dict[str, str]
DllErrMsgDict = dict[str, FuncErrMsgDict]
DriveList = list[tuple[Path, Path]]

c_ssize_p = POINTER(c_ssize_t)
UCHAR = c_ubyte
PUCHAR = POINTER(UCHAR)
PUCHARType = type[PUCHAR]
ULONGLONG = c_ulonglong

MAX_PREFERRED_LENGTH = -1

ERROR_INVALID_HANDLE = 6

PROCESS_DPI_UNAWARE = 0
PROCESS_SYSTEM_DPI_AWARE = 1
PROCESS_PER_MONITOR_DPI_AWARE = 2

WIN_WCHAR_ENCODING = 'utf-16-le'
"""
Win API only has UTF-16-LE support because they were early adopters before
the rest of the internet converged on utf-8.  Calls to Win API funcs need to
encode as utf-16-le.  Use this constant when doing low-level codec operations.
"""

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/596a1078-e883-4972-9bbc-49e60bebca55  # noqa: E501
NTSTATUS_CODES = {
    0x00000000: 'STATUS_SUCCESS',
    0xC0000225: 'STATUS_NOT_FOUND',
    0xC000000D: 'STATUS_INVALID_PARAMETER',
    0xC0000017: 'STATUS_NO_MEMORY',
    0xC0000008: 'STATUS_INVALID_HANDLE',
    0xC00000BB: 'STATUS_NOT_SUPPORTED',
    0xC0000023: 'STATUS_BUFFER_TOO_SMALL',
    0xC000A002: 'STATUS_AUTH_TAG_MISMATCH',
    0xC0000206: 'STATUS_INVALID_BUFFER_SIZE',
}
